import heapq
import logging
import signal
import time
from collections.abc import Callable
from concurrent.futures import Future, ThreadPoolExecutor
from datetime import datetime, timezone
from threading import Lock
from typing import List, Tuple

from .task import Task


class Scheduler:
    _loop_delay: float = 0.01

    def __init__(self, *tasks: Task):
        self._stopped = False
        self._tasks = tasks
        self._lock = Lock()

        assert len(tasks) > 0
        assert len(tasks) == len(
            set(task.get_name() for task in tasks)
        ), 'Task names should be unique'

        self._executor = ThreadPoolExecutor(
            max_workers=len(tasks),
            thread_name_prefix='scheduler',
        )
        self._tasks_q: List[Tuple[datetime, str, Task]] = []
        self._prepare_tasks(*tasks)
        self._futures_mapper = {}

        self._logger = logging.getLogger('scheduler_mapper')

        self._register_sigterm_handler()

    def _register_sigterm_handler(self):

        def new_handler(*args, **kwargs):
            self.stop()

            if isinstance(old_handler, Callable):
                old_handler(*args, **kwargs)

        old_handler = signal.getsignal(signal.SIGTERM)
        signal.signal(signal.SIGTERM, new_handler)

    def stop(self):
        self._logger.info('Trying to gracefully stop scheduler')
        self._stopped = True
        self._logger.info('Waiting for all pending tasks...')
        self._executor.shutdown(wait=True)
        self._logger.info('Stopping done')

    @property
    def is_stopped(self):
        return self._stopped

    def run(self):
        while not self._stopped:
            try:
                self._run()
            except (KeyboardInterrupt, SystemExit):
                self.stop()
                continue
            except Exception:
                self._logger.exception(
                    'Unexpected error occurred, trying again'
                )
                time.sleep(self._loop_delay)
                continue

    def _run(self):
        while not self._stopped:
            if len(self._tasks_q) == 0:
                time.sleep(self._loop_delay)
                continue

            with self._lock:
                next_task = self._peak_nearest_task()
                next_run = next_task.get_next_run_datepoint()

                utcnow = datetime.now(timezone.utc)
                next_run_delay = (next_run - utcnow).total_seconds()

                if next_run_delay > 0:
                    time.sleep(self._loop_delay)
                    continue

                self._pop_nearest_task()
            self._run_task(next_task)

    def _peak_nearest_task(self) -> Task:
        return self._tasks_q[0][2]

    def _pop_nearest_task(self):
        heapq.heappop(self._tasks_q)

    def _push_task(self, task: Task):
        with self._lock:
            heapq.heappush(
                self._tasks_q, (
                    task.get_next_run_datepoint(),
                    task.get_uniq_name(),
                    task,
                )
            )

    def _prepare_tasks(self, *tasks: Task):
        for task in tasks:
            self._update_task_schedule(task)

    def _on_task_done(self, future: Future):
        task = self._futures_mapper[id(future)]
        self._update_task_schedule(task)

    def _update_task_schedule(self, task: Task):
        self._push_task(task)

    def _run_task(self, task: Task) -> None:
        # Это случается, если прилетает sigterm
        if self._stopped:
            self._logger.info('Scheduler is stopped, cancel task')
            return

        job = task.get_job()

        future = self._executor.submit(job)
        self._futures_mapper[id(future)] = task
        future.add_done_callback(self._on_task_done)



import logging
import uuid
from datetime import datetime, timezone
from typing import Any, Callable, Dict, Hashable, Optional, Tuple

from croniter import croniter


class Task:

    def __init__(
        self,
        name: str,
        cron_schedule: str,
        job: Callable,
        job_args: Optional[Tuple[Any, ...]] = None,
        job_kwargs: Optional[Dict[Hashable, Any]] = None,
        *,
        is_overdue_gap_needed: bool = True,
    ):
        self._name = name
        self._instance_key = uuid.uuid4()
        self._job = job

        # Не рекомендуется использовать мутирующие объекты в качестве аргументов
        self._job_args = job_args or ()
        self._job_kwargs = job_kwargs or {}
        self._cron_schedule = cron_schedule

        self._croniter = self._create_calendar()
        self._is_overdue_gap_needed = is_overdue_gap_needed

        self._next_run = self._croniter.get_next(datetime)
        self._last_started_at = None

        self._logger = logging.getLogger(f'{"TODO"}.task')  #TODO

    def __str__(self) -> str:
        return (
            f'Task with job [{repr(self._job)}] '
            f'and cron schedule [{self._cron_schedule}]'
        )

    def get_name(self) -> str:
        return self._name

    def get_uniq_name(self) -> str:
        return f'{self._name}_{self._instance_key}'

    def get_job(self) -> Callable:
        return self._run

    def get_next_run_datepoint(self) -> datetime:
        return self._next_run

    def _run(self):
        self._logger.info(
            'Task [%s] started with schedule [%s]',
            self._name,
            self._cron_schedule,
        )

        self._last_started_at = datetime.now(timezone.utc)
        try:
            self._job(*self._job_args, **self._job_kwargs)
        except Exception:
            self._logger.exception(
                'Unexpected error occurred in task [%s]',
                self._name,
            )
        else:
            self._logger.info(
                'Task completed [%s]',
                self._name,
            )

        self._on_finish()

    def _on_finish(self):
        finished_at = datetime.now(timezone.utc)

        optimistic_next_run = self._croniter.get_next(
            datetime,
            self._croniter.get_current(),
        )

        is_overdue = finished_at >= optimistic_next_run
        if not is_overdue:
            self._next_run = optimistic_next_run

        if is_overdue and self._is_overdue_gap_needed:
            self._croniter.set_current(finished_at)
            self._next_run = self._croniter.get_next(datetime)

        if is_overdue and not self._is_overdue_gap_needed:
            self._croniter.set_current(finished_at)
            self._next_run = finished_at

    def _create_calendar(self) -> croniter:
        assert croniter.is_valid(self._cron_schedule), 'No valid cron schedule'
        return croniter(self._cron_schedule, datetime.now(timezone.utc))



import setuptools

# Reads the content of your README.md into a variable to be used in the setup below
with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name='ssd-scheduler',  # should match the package folder
    packages=['ssd-scheduler'],  # should match the package folder
    version='v1.2',  # important for updates
    license='MIT',  # should match your chosen license
    description='Testing installation of Package',
    long_description=long_description,  # loads your README.md
    long_description_content_type="text/markdown",  # README.md is of type 'markdown'
    author='Test user',
    author_email='test@gmail.com',
    url='https://murathanov-stepan.ru/gitea/sa/ssd-scheduler.git',
    project_urls={  # Optional
        "Bug Tracker": "https://murathanov-stepan.ru/gitea/sa/ssd-scheduler.git"
    },
    install_requires=['requests'],  # list all packages that your package uses
    keywords=["pypi", "ssd-scheduler", "test"],  # descriptive meta-data
    classifiers=[  # https://pypi.org/classifiers
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Documentation',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ],

    download_url="https://murathanov-stepan.ru/gitea/sa/ssd-scheduler/archive/v1.2.tar.gz",
)
